@@ -1,4 +1,4 @@
-# Copyright 2020 The TensorFlow Recommenders Authors.
+# Copyright 2021 The TensorFlow Recommenders Authors.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -21,13 +21,16 @@ It's built on Keras and aims to have a gentle learning curve while still giving
 you the flexibility to build complex models.
 """
 
-__version__ = "v0.3.1"
+__version__ = "v0.5.0"
 
 from tensorflow_recommenders import examples
+from tensorflow_recommenders import experimental
 # Internal extension library import.
 from tensorflow_recommenders import layers
 from tensorflow_recommenders import metrics
 from tensorflow_recommenders import models
 from tensorflow_recommenders import tasks
+from tensorflow_recommenders import types
+
 
 Model = models.Model
